clear all; close all;

%% (a) Load images

image1 = imread('../data/fountain/0005.png');
image2 = imread('../data/fountain/0007.png');

image1 = rgb2gray(image1);
image2 = rgb2gray(image2);

C = [2759.48 0 1520.69 
     0 2764.16 1006.81 
     0 0 1]; 
 
 
%% (b) Find point pairs using keypoint detectionw

% Extract FAST and SURF keypoints
corner_fast_1 = detectFASTFeatures(image1, 'MinContrast', 0.08);
corner_fast_2 = detectFASTFeatures(image2, 'MinContrast', 0.08);
blob_surf_1 = detectSURFFeatures(image1);
blob_surf_2 = detectSURFFeatures(image2);

% figure; imshow(image1); hold on;
% plot(corner_fast_1.selectStrongest(200));
% figure; imshow(image2); hold on;
% plot(corner_fast_2.selectStrongest(200));
% 
% figure; imshow(image1); hold on;
% plot(blob_surf_1.selectStrongest(200));
% figure; imshow(image2); hold on;
% plot(blob_surf_2.selectStrongest(200));

% Extract BRISK and SURF descriptors
% FAST-BRISK
[fast_brisk_1, valid_fast_brisk_1] = extractFeatures(image1, corner_fast_1, 'Method', 'BRISK');
[fast_brisk_2, valid_fast_brisk_2] = extractFeatures(image2, corner_fast_2, 'Method', 'BRISK');

% FAST-SURF
[fast_surf_1, valid_fast_surf_1] = extractFeatures(image1, corner_fast_1, 'Method', 'SURF');
[fast_surf_2, valid_fast_surf_2] = extractFeatures(image2, corner_fast_2, 'Method', 'SURF');

% SURF-BRISK
[surf_brisk_1, valid_surf_brisk_1] = extractFeatures(image1, blob_surf_1, 'Method', 'BRISK');
[surf_brisk_2, valid_surf_brisk_2] = extractFeatures(image2, blob_surf_2, 'Method', 'BRISK');

% SURF-SURF
[surf_surf_1, valid_surf_surf_1] = extractFeatures(image1, blob_surf_1, 'Method', 'SURF');
[surf_surf_2, valid_surf_surf_2] = extractFeatures(image2, blob_surf_2, 'Method', 'SURF');

% figure; imshow(image1); hold on;
% plot(valid_fast_brisk_1.selectStrongest(200));
% figure; imshow(image2); hold on;
% plot(valid_fast_brisk_2.selectStrongest(200));
% 
% figure; imshow(image1); hold on;
% plot(valid_fast_surf_1.selectStrongest(200));
% figure; imshow(image2); hold on;
% plot(valid_fast_surf_2.selectStrongest(200));
% 
% figure; imshow(image1); hold on;
% plot(valid_surf_brisk_1.selectStrongest(200), 'showOrientation',true);
% figure; imshow(image2); hold on;
% plot(valid_surf_brisk_2.selectStrongest(200), 'showOrientation',true);
% 
% figure; imshow(image1); hold on;
% plot(valid_surf_surf_1.selectStrongest(200), 'showOrientation',true);
% figure; imshow(image2); hold on;
% plot(valid_surf_surf_2.selectStrongest(200), 'showOrientation',true);

% Matching using the second best ratio distance

% Matching using SSD or Hamming distance accordingly
% FAST-BRISK
fast_brisk_Pairs = matchFeatures(fast_brisk_1, fast_brisk_2) ;
fast_brisk_matchedPoints1 = valid_fast_brisk_1(fast_brisk_Pairs(:,1));
fast_brisk_matchedPoints2 = valid_fast_brisk_2(fast_brisk_Pairs(:,2));
figure; showMatchedFeatures(image1, image2, fast_brisk_matchedPoints1, fast_brisk_matchedPoints2);
legend('matched points 1','matched points 2');

% FAST-SURF
fast_surf_Pairs = matchFeatures(fast_surf_1, fast_surf_2) ;
fast_surf_matchedPoints1 = valid_fast_surf_1(fast_surf_Pairs(:,1));
fast_surf_matchedPoints2 = valid_fast_surf_2(fast_surf_Pairs(:,2));
figure; showMatchedFeatures(image1, image2, fast_surf_matchedPoints1, fast_surf_matchedPoints2);
legend('matched points 1','matched points 2');

% SURF-BRISK
surf_brisk_Pairs = matchFeatures(surf_brisk_1, surf_brisk_2) ;
surf_brisk_matchedPoints1 = valid_surf_brisk_1(surf_brisk_Pairs(:,1));
surf_brisk_matchedPoints2 = valid_surf_brisk_2(surf_brisk_Pairs(:,2));
figure; showMatchedFeatures(image1, image2, surf_brisk_matchedPoints1, surf_brisk_matchedPoints2);
legend('matched points 1','matched points 2');

% SURF-SURF
surf_surf_Pairs = matchFeatures(surf_surf_1, surf_surf_2) ;
surf_surf_matchedPoints1 = valid_surf_surf_1(surf_surf_Pairs(:,1));
surf_surf_matchedPoints2 = valid_surf_surf_2(surf_surf_Pairs(:,2));
figure; showMatchedFeatures(image1, image2, surf_surf_matchedPoints1, surf_surf_matchedPoints2);
legend('matched points 1','matched points 2');


% % Matching using second best ratio distance accordingly
% % SURF-BRISK
% distances_surf_brisk = pdist2(surf_brisk_1.Features, surf_brisk_2.Features, @bitwiseHammingDist);
% [~, argmin_surf_brisk] = min(distances_surf_brisk, [], 2);
% distances_surf_brisk = sort(distances_surf_brisk, 2);
% distances_surf_brisk = distances_surf_brisk(:, 1)./distances_surf_brisk(:, 2);
% 
% surf_brisk_Pairs2 = [(1:size(argmin_surf_brisk, 1))' argmin_surf_brisk];
% [distances_surf_brisk, ordered_pair_id] = sort(distances_surf_brisk);
% surf_brisk_matchedPoints1_sb = valid_surf_brisk_1(surf_brisk_Pairs2(ordered_pair_id(1:50),1));
% surf_brisk_matchedPoints2_sb = valid_surf_brisk_2(surf_brisk_Pairs2(ordered_pair_id(1:50),2));
% figure; showMatchedFeatures(image1, image2, surf_brisk_matchedPoints1_sb, surf_brisk_matchedPoints2_sb);
% legend('matched points 1','matched points 2');
% 
% % SURF-SURF
% distances_surf_surf = pdist2(surf_surf_1, surf_surf_2, 'squaredeuclidean');
% [~, argmin_surf_surf] = min(distances_surf_surf, [], 2);
% distances_surf_surf = sort(distances_surf_surf, 2);
% distances_surf_surf = distances_surf_surf(:, 1)./distances_surf_surf(:, 2);
% 
% surf_surf_Pairs2 = [(1:size(argmin_surf_surf, 1))' argmin_surf_surf];
% [distances_surf_surf, ordered_pair_id2] = sort(distances_surf_surf);
% surf_surf_matchedPoints1_sb = valid_surf_surf_1(surf_surf_Pairs2(ordered_pair_id2(1:40),1));
% surf_surf_matchedPoints2_sb = valid_surf_surf_2(surf_surf_Pairs2(ordered_pair_id2(1:40),2));
% figure; showMatchedFeatures(image1, image2, surf_surf_matchedPoints1_sb, surf_surf_matchedPoints2_sb);
% legend('matched points 1','matched points 2');


%% (c) 2D-to-2D motion estimation using RANSAC based 8-point-algorithm

% Success probability
p = 0.99;
% Outlier ratio
epsilon = 0.5;
% Threshold for inlier
r = 10;
r = 1;
% Number of required data points
s = 8;
%Number of iterations
N = uint8(log(1 - p) / log(1 - (1 - epsilon)^s));
% Number of matched pairs
n = size(surf_surf_Pairs, 1);

best_R = zeros(3);
best_T = zeros(3, 1);
best_num_inliers = -1;

for i = 1:N
   % Randomly select s data points
   index = randsample(n, s);
   sampled_matched_points1 = valid_surf_surf_1(surf_surf_Pairs(index, 1));
   sampled_matched_points2 = valid_surf_surf_2(surf_surf_Pairs(index, 2));
   
   % Caculate R and T from the s points
   [R, T] = eight_point_algorithm(sampled_matched_points1.Location(:, 1),...
                                  sampled_matched_points1.Location(:, 2),...
                                  sampled_matched_points2.Location(:, 1),...
                                  sampled_matched_points2.Location(:, 2),...
                                  C);
   
   % Reconstruct 3D points
   [~, p] = reconstruction(R, T, ...
                           surf_surf_matchedPoints1.Location(:, 1),...
                           surf_surf_matchedPoints1.Location(:, 2),...
                           surf_surf_matchedPoints2.Location(:, 1),...
                           surf_surf_matchedPoints2.Location(:, 2),...
                           n, C);
    
   % Count inliers
   p2 = (R * p' + repmat(T, 1, size(p', 2)))';
   p2 = p2 ./ repmat(p2(:, 3), 1, 3);
   image_p2 = (C * p2')';
   diff = [image_p2(:, 1) - surf_surf_matchedPoints2.Location(:, 1) ...
           image_p2(:, 2) - surf_surf_matchedPoints2.Location(:, 2)];
   errors = diff(:, 1).^2 + diff(:, 2).^2;
   indices = errors < r * r;
   num_inliers = sum(indices);
   
   % Keep best
   if (num_inliers > best_num_inliers)
       best_num_inliers = num_inliers;
       best_R = R;
       best_T = T;
       best_inliers_indices = indices;
   end
end

% Refit model using inliers
inlier_pairs = surf_surf_Pairs(best_inliers_indices, :);
inlier_surf_surf_matchedPoints1 = valid_surf_surf_1(inlier_pairs(:,1));
inlier_surf_surf_matchedPoints2 = valid_surf_surf_2(inlier_pairs(:,2));
figure; showMatchedFeatures(image1, image2, inlier_surf_surf_matchedPoints1, inlier_surf_surf_matchedPoints2);
legend('matched points 1','matched points 2');

best_R
best_T
[R, T] = eight_point_algorithm(inlier_surf_surf_matchedPoints1.Location(:, 1),...
                               inlier_surf_surf_matchedPoints1.Location(:, 2),...
                               inlier_surf_surf_matchedPoints2.Location(:, 1),...
                               inlier_surf_surf_matchedPoints2.Location(:, 2),...
                               C)

                           
